/**********************************************************************
 *<
	FILE: particlectrl.cpp

	DESCRIPTION: A controller that moves an object along a spline particle
				 Ed. 2 re-coded using ParamBlock2/ParamMap2

	CREATED BY: Rolf Berteig
	            Ed. 2 John Wainwright

	HISTORY: created 13 June 1995
	         Ed. 2 re-coded 9/22/98

 *>	Copyright (c) 1994, All Rights Reserved.
 **********************************************************************/

#include "ctrl.h"
#include "units.h"
#include "interpik.h"
#include "spline3d.h"
#include "istdplug.h"
#include "iparamm2.h"
#include "simpobj.h"

#define PARTICLE_CONTROL_CLASS_ID 0x4f676666
#define PARTICLE_CONTROL_CNAME		GetString(IDS_RB_PARTICLE)

#define InterpType() ((ConstVel())?SPLINE_INTERP_NORMALIZED:SPLINE_INTERP_SIMPLE)

// flag values
#define PARTICLEFLAG_FOLLOW		(1<<0)
#define PARTICLEFLAG_BANK		(1<<1)
#define PARTICLEFLAG_CLOSED		(1<<2)
#define PARTICLEFLAG_ALLOWFLIP	(1<<3)
#define PARTICLEFLAG_CONSTVEL	(1<<4)
#define PARTICLEFLAG_FLIP		(1<<5)
#define PARTICLEFLAG_YAXIS		(1<<6)

#define PARTICLEFLAG_ZAXIS		(1<<7)


#define PARTICLEPOS_PBLOCK_REF 0

// ParamBlock IDs
// main block
enum { particle_params};
// particle_params param IDs
enum { particle_ID, particle_pick, particle_follow, particle_bank, particle_bank_amount, particle_smoothness, 
       particle_allow_upsidedown, particle_axis, particle_axis_flip, particle_inc};


class ParticlePosition : public	Control {
	public:
		IParamBlock2* pblock;
//		INode *particle;
		DWORD flags;
		float bankAmount, tracking;

		BOOL incClone;

		Point3 curval;
		Quat curRot;
		Interval ivalid;
		
		static ParticlePosition *editCont;
		static IObjParam *ip;

		ParticlePosition(const ParticlePosition &ctrl);
		ParticlePosition(BOOL loading=FALSE);
		~ParticlePosition();

//		BOOL SetParticleNode(INode *node);
		void SetBankAmount(float a);
		void SetTracking(float t);
		void SetFollow(BOOL f);
		void SetBank(BOOL b);
		void SetAllowFlip(BOOL f);
//		void SetConstVel(BOOL cv);
		void SetFlip(BOOL onOff);
		void SetAxis(int axis);
		
		BOOL GetFollow() {return Follow();}
		float GetBankAmount();
		BOOL GetBank() {return Bank();}
		float GetTracking();
		BOOL GetAllowFlip() {return AllowFlip();}
//		BOOL GetConstVel() {return ConstVel();}		

		BOOL Follow() {return pblock->GetInt(particle_follow, 0);}
		BOOL Bank() {return pblock->GetInt(particle_bank, 0);}
		BOOL AllowFlip() {return pblock->GetInt(particle_allow_upsidedown, 0);}
//		BOOL ConstVel() {return flags&PARTICLEFLAG_CONSTVEL?TRUE:FALSE;}
		BOOL GetFlip() {return pblock->GetInt(particle_axis_flip, 0);}
		int GetAxis();		
		
		void Update(TimeValue t);
		int GetID(TimeValue t,Interval &valid,BOOL noClip=FALSE);
//		Point3 PointOnParticle(TimeValue t, ShapeObject *particleOb, Interval &valid);
//		Matrix3 CalcRefFrame(TimeValue t,ParticleObject *particleOb);
		Matrix3 CalcRefFrame(TimeValue t,ParticleObject *pobj);

		// Animatable methods
		Class_ID ClassID() { return Class_ID(PARTICLE_CONTROL_CLASS_ID,0); }  
		SClass_ID SuperClassID() { return CTRL_POSITION_CLASS_ID; }  		
		
		void GetClassName(TSTR& s) {s = PARTICLE_CONTROL_CNAME;}
		void DeleteThis() {delete this;}		
		int IsKeyable() {return 0;}		

		int NumSubs()  {return 1;}
		Animatable* SubAnim(int i) {return pblock;}
		TSTR SubAnimName(int i) { return GetString(IDS_RB_PARTICLEPARAMS); }
		int SubNumToRefNum(int subNum) {if (subNum==0) return PARTICLEPOS_PBLOCK_REF; else return -1;}

		void BeginEditParams( IObjParam *ip, ULONG flags,Animatable *prev ); 
		void EndEditParams( IObjParam *ip, ULONG flags,Animatable *next ); 

		int SetProperty(ULONG id, void *data);
		void *GetProperty(ULONG id);

// JBW: direct ParamBlock access is added
		int	NumParamBlocks() { return 1; }					// return number of ParamBlocks in this instance
		IParamBlock2* GetParamBlock(int i) { return pblock; } // return i'th ParamBlock
		IParamBlock2* GetParamBlockByID(BlockID id) { return (pblock->ID() == id) ? pblock : NULL; } // return id'd ParamBlock

		// Reference methods
		int NumRefs() { return 1; };	
		RefTargetHandle GetReference(int i);
		void SetReference(int i, RefTargetHandle rtarg);
		RefResult NotifyRefChanged(Interval, RefTargetHandle, PartID&, RefMessage);

		IOResult Save(ISave *isave);
		IOResult Load(ILoad *iload);

		void RescaleWorldUnits(float f) {}

		// Control methods
		void Copy(Control *from);
		RefTargetHandle Clone(RemapDir& remap);
		BOOL IsLeaf() {return FALSE;}
		void GetValue(TimeValue t, void *val, Interval &valid, GetSetMethod method);	
		void SetValue(TimeValue t, void *val, int commit, GetSetMethod method) {}
		void CommitValue(TimeValue t) { }
		void RestoreValue(TimeValue t) { }
	};
/*
class JointParamsParticle : public JointParams {
	public:			 	
		JointParamsParticle() : JointParams((DWORD)JNT_POS,1,100.0f) {}
		void SpinnerChange(InterpCtrlUI *ui,WORD id,ISpinnerControl *spin,BOOL interactive);
	};
*/
IObjParam *ParticlePosition::ip             = NULL;
ParticlePosition *ParticlePosition::editCont    = NULL;

//********************************************************
// PARTICLE CONTROL
//********************************************************
static Class_ID particleControlClassID(PARTICLE_CONTROL_CLASS_ID,0); 
class ParticleControlClassDesc:public ClassDesc2 {
	public:
	int 			IsPublic() { return 1; }
	void *			Create(BOOL loading) { return new ParticlePosition(loading); }
	const TCHAR *	ClassName() { return PARTICLE_CONTROL_CNAME; }
	SClass_ID		SuperClassID() { return CTRL_POSITION_CLASS_ID; }
	Class_ID		ClassID() { return particleControlClassID; }
	const TCHAR* 	Category() { return _T("");  }
// JBW: new descriptor data accessors added.
	const TCHAR*	InternalName() { return _T("particleControl"); }			// returns fixed parsable name (scripter-visible name)
	HINSTANCE		HInstance() { return hInstance; }				// returns owning module handle
	};

static ParticleControlClassDesc particleCD;
ClassDesc* GetParticleCtrlDesc() {return &particleCD;}


// parameter setter callback, reflect any ParamBlock-mediated param setting in instance data members.
// JBW: since the old particle controller kept all parameters as instance data members, this setter callback
// is implemented to to reduce changes to existing code 
class ParticlePBAccessor : public PBAccessor
{
public:
	void Set(PB2Value& v, ReferenceMaker* owner, ParamID id, int tabIndex)    // set from v
	{
		ParticlePosition* p = (ParticlePosition*)owner;
		switch (id)
		{
			case particle_follow:
				if (v.i) p->flags |= PARTICLEFLAG_FOLLOW; else p->flags &= ~PARTICLEFLAG_FOLLOW; break;
			case particle_bank:
				if (v.i) p->flags |= PARTICLEFLAG_BANK; else p->flags &= ~PARTICLEFLAG_BANK; break;
			case particle_bank_amount:
				p->bankAmount = FromBankUI(v.f); break;
			case particle_smoothness:
				p->tracking = FromTrackUI(v.f); break;
			case particle_allow_upsidedown:
				if (v.i) p->flags |= PARTICLEFLAG_ALLOWFLIP; else p->flags &= ~PARTICLEFLAG_ALLOWFLIP; break;
/*			case particle_constant_vel:
				if (v.i) p->flags |= PARTICLEFLAG_CONSTVEL; else p->flags &= ~PARTICLEFLAG_CONSTVEL; break;
*/
			case particle_axis:
				switch (v.i)
				{
					case 0: p->flags &= ~(PARTICLEFLAG_YAXIS | PARTICLEFLAG_ZAXIS); break;			// X
					case 1: p->flags |= PARTICLEFLAG_YAXIS; p->flags &= ~PARTICLEFLAG_ZAXIS; break;	// Y
					case 2: p->flags |= PARTICLEFLAG_ZAXIS; p->flags &= ~PARTICLEFLAG_YAXIS; break;	// Y
				}
				break;
			case particle_axis_flip:
				if (v.i) p->flags |= PARTICLEFLAG_FLIP; else p->flags &= ~PARTICLEFLAG_FLIP; break;
		}
	}
};

static ParticlePBAccessor particle_accessor;

class ParticleValidatorClass : public PBValidator
{
BOOL Validate(PB2Value &v) 
	{
	INode *node = (INode*) v.r;
	const ObjectState& os = node->EvalWorldState(0);
	Object* ob = os.obj;
	if (ob && ob->IsParticleSystem()) {
		return TRUE;
//		ParticleObject *obj = (ParticleObject*)ob;
		}

	return FALSE;

	};
};

static ParticleValidatorClass  ParticleValidator;

// per instance particle controller block
static ParamBlockDesc2 particle_paramblk (particle_params, _T("ParticleParameters"),  0, &particleCD, P_AUTO_CONSTRUCT + P_AUTO_UI, PARTICLEPOS_PBLOCK_REF, 
	//rollout
	IDD_PARTICLEPARAMS, IDS_RB_PARTICLEPARAMS, BEGIN_EDIT_MOTION, 0, NULL,
	// params
	particle_ID,	_T("particleID"),		TYPE_INT, P_ANIMATABLE + P_RESET_DEFAULT, 	IDS_RB_ID, 
		p_default, 		0,	
		p_range, 		0, 999999999, 
		p_ui, 			TYPE_SPINNER, EDITTYPE_INT, IDC_PARTICLEID, IDC_PARTICLEIDSPIN, 1.0f, 
		end, 
	particle_pick, 		_T("particle"), 		TYPE_INODE, 	0,	IDS_RB_PARTICLE,
		p_ui, 			TYPE_PICKNODEBUTTON, IDC_PICKPARTICLE, 
		p_validator,    &ParticleValidator,
//		p_sclassID,		SHAPE_CLASS_ID, 
		end, 
//	particlemesher_pick,  _T("pick"), 	TYPE_INODE, 	0, 	IDS_PW_PICK, 
//		p_ui, 			TYPE_PICKNODEBUTTON,  IDC_PICK,
//		end, 

	particle_follow,	_T("follow"),		TYPE_BOOL, 		0,									IDS_JW_FOLLOW,
		p_default, 		FALSE, 
		p_ui, 			TYPE_SINGLECHEKBOX, IDC_PARTICLE_FOLLOW, 
		p_enable_ctrls,	4, particle_bank, particle_allow_upsidedown, particle_axis, particle_axis_flip,
		p_accessor,		&particle_accessor,
		end, 
	particle_bank,		_T("bank"),			TYPE_BOOL, 		0,									IDS_JW_BANK,
		p_default, 		FALSE, 
		p_ui, 			TYPE_SINGLECHEKBOX, IDC_PARTICLE_BANK, 
		p_enabled,		FALSE,
		p_enable_ctrls,	2, particle_bank_amount, particle_smoothness,
		p_accessor,		&particle_accessor,
		end, 
	particle_bank_amount, _T("bankAmount"), TYPE_FLOAT, 	P_ANIMATABLE, 						IDS_JW_BANKAMOUNT, 
		p_default, 		0.5, 
		p_range, 		-999999.0, 999999.0, 
		p_ui, 			TYPE_SPINNER, EDITTYPE_FLOAT, IDC_FOLLOW_BANK, IDC_FOLLOW_BANKSPIN, 0.01f,
		p_accessor,		&particle_accessor,
		p_enabled,		FALSE,
		end, 
	particle_smoothness,  _T("smoothness"), TYPE_FLOAT, 	P_ANIMATABLE, 						IDS_JW_SMOOTHNESS, 
		p_default, 		0.5, 
		p_range, 		0.01f, 10.0f, 
		p_ui, 			TYPE_SPINNER, EDITTYPE_FLOAT, IDC_FOLLOW_TRACK, IDC_FOLLOW_TRACKSPIN, 0.01f,
		p_enabled,		FALSE,
		p_accessor,		&particle_accessor,
		end, 
	particle_allow_upsidedown, _T("allowUpsideDown"), TYPE_BOOL, 0,								IDS_JW_ALLOWUPSIDEDOWN,
		p_default, 		FALSE, 
		p_ui, 			TYPE_SINGLECHEKBOX, IDC_PARTICLE_ALLOWFLIP, 
		p_enabled,		FALSE,
		p_accessor,		&particle_accessor,
		end, 
	particle_axis, 		_T("axis"),			TYPE_INT, 		0,									IDS_JW_AXIS,
		p_default, 		0, 
		p_range, 		0, 2, 
		p_ui, 			TYPE_RADIO, 	3, IDC_PARTICLE_X, IDC_PARTICLE_Y, IDC_PARTICLE_Z, 
		p_enabled,		FALSE,
		p_accessor,		&particle_accessor,
		end, 
	particle_axis_flip, _T("axisFlip"),		TYPE_BOOL,		0,									IDS_JW_AXISFLIP,
		p_default, 		FALSE, 
		p_ui, 			TYPE_SINGLECHEKBOX, IDC_PARTICLE_FLIP, 
		p_enabled,		FALSE,
		p_accessor,		&particle_accessor,
		end, 

	particle_inc,	_T("incrementOnClone"),		TYPE_BOOL, 		0,									IDS_JW_FOLLOW,
		p_default, 		TRUE, 
		p_ui, 			TYPE_SINGLECHEKBOX, IDC_PARTICLE_INCREMENT, 
		end, 

	end
	);

ParticlePosition::ParticlePosition(BOOL loading) 
	{
	curval   = Point3(0,0,0);
	curRot.Identity();
	flags      = 0;
	bankAmount = FromBankUI(0.5f);
	tracking   = FromTrackUI(0.5f);
	incClone = TRUE;

	// make the paramblock
	particleCD.MakeAutoParamBlocks(this);
	// pre-animate percent param
/*	SuspendAnimate();
	AnimateOn();
	float val =  1.0f;
	pblock->SetValue(particle_percent, GetAnimEnd(), val); 
	ResumeAnimate();
*/
	ivalid.SetEmpty();
}

ParticlePosition::~ParticlePosition()
	{
	DeleteAllRefsFromMe();
	}

RefTargetHandle ParticlePosition::Clone(RemapDir& remap)
	{
	ParticlePosition *p = new ParticlePosition(TRUE);

/*	if (particle) {
		p->ReplaceReference(PARTICLEPOS_PARTICLE_REF,particle);
	} else {
		p->particle = NULL;
		}
*/
    p->ReplaceReference(PARTICLEPOS_PBLOCK_REF, pblock->Clone(remap));

	p->flags      = flags;
	p->bankAmount = bankAmount;
	p->tracking   = tracking;
	p->curval     = curval;
	p->curRot     = curRot;
	p->incClone = incClone;
	int id = GetID(0,FOREVER,TRUE)+1;
	if (incClone)
		p->pblock->SetValue(particle_ID, 0,id );
	p->ivalid.SetEmpty();
	BaseClone(this, p, remap);
	return p;
	}


void ParticlePosition::Copy(Control *from)
	{
	if (from->ClassID()==ClassID()) {
		ParticlePosition *ctrl = (ParticlePosition*)from;
//		ReplaceReference(PARTICLEPOS_PARTICLE_REF,ctrl->particle);
		// a copy will construct its own pblock to keep the pblock-to-owner 1-to-1.
		RemapDir *remap = NewRemapDir(); 
		ReplaceReference(PARTICLEPOS_PBLOCK_REF, ctrl->pblock->Clone(*remap));
		remap->DeleteThis();
		curval   = ctrl->curval;
		curRot   = ctrl->curRot;
		ivalid   = ctrl->ivalid;		
	} else {
		from->GetValue(0,&curval,ivalid);
		}
	NotifyDependents(FOREVER,PART_ALL,REFMSG_CHANGE);
	}

int ParticlePosition::GetID(TimeValue t,Interval &valid,BOOL noClip)
	{	
	int id;
	pblock->GetValue(particle_ID, t, id, valid);	
	return id;
	}


// Reference frame is calculated as the following:
//
// X axis = tangent(t)
// Y axis = WorldZ cross X
// Z axis = X cross Y
//
// or if the 'Allow Upside Down' option us selected:
//
// X axis = tangent(t)
// Z axis = particle normal cross X
// Y axis = Z cross X
//
// If roll is on then the reference frame is rotated about
// the tangent by rollamount * curviture.

#define DU			0.001f
#define BDU			0.01f
#define NUMSTEPS	5

//Matrix3 ParticlePosition::CalcRefFrame(TimeValue t,ParticleObject *particleOb)
Matrix3 ParticlePosition::CalcRefFrame(TimeValue t,ParticleObject *pobj)
	{
	Interval valid;
	//INode *particle;
	//pblock->GetValue(particle_particle,t,particle,FOREVER);
	Matrix3 tm(1);//, otm = particle->GetObjTMAfterWSM(t);
//	int u = GetID(t,valid);
	bankAmount = FromBankUI(pblock->GetFloat(particle_bank_amount, t));
	tracking = FromTrackUI(pblock->GetFloat(particle_smoothness, t));
	
	Point3 pt0, pt1;
//	pt0 = particleOb->InterpCurve3D(t, 0, u-DU, InterpType()) * otm;
//get velocity
//	pt1 = particleOb->InterpCurve3D(t, 0, u+DU, InterpType()) * otm;
	int id;
	pblock->GetValue(particle_ID,t,id,FOREVER);


	pt0 = pobj->ParticlePosition(t,id);
	pt1 = pt0+pobj->ParticleVelocity(t,id);

	TimeValue age = pobj->ParticleAge(t,id);
	TimeValue life = pobj->ParticleLife(t,id);

	if ((age > life) || (age < 0))
		{
		pt0.x = 0.0f;
		pt0.y = 0.0f;
		pt0.z = 0.0f;
		pt1.x = 0.0f;
		pt1.y = 0.0f;
		pt1.z = 1.0f;
		}
	
	Point3 particleNorm(0,0,0);
#define NUM_SAMPLES 20

	TimeValue StartTime = t - GetTicksPerFrame()/2;
	TimeValue Inc =  GetTicksPerFrame()/NUM_SAMPLES;
	if (AllowFlip()) {
		// Compute the normal to the plane of the particle by sampling points on the particle

		Point3 v[NUM_SAMPLES], cent(0,0,0);
		TimeValue st = StartTime;
		for (int i=0; i<NUM_SAMPLES; i++) {
//			v[i] = particleOb->InterpCurve3D(t, 0, float(i)/float(NUM_SAMPLES), InterpType()) * otm;
			v[i] = pobj->ParticlePosition(st,id);
			st += Inc;
			cent += v[i];
			}		
		cent /= float(NUM_SAMPLES);
		for (i=1; i<NUM_SAMPLES; i++) {
			particleNorm += Normalize((v[i]-cent)^(v[i-1]-cent));
			}

		particleNorm = Normalize(particleNorm);
		}

	// X
	tm.SetRow(0,Normalize(pt1-pt0));

	if (GetFlip()) tm.SetRow(0,-tm.GetRow(0));

	if (AllowFlip()) {
		// Choose Z in the plane of the particle		
		// Z
		tm.SetRow(2,Normalize(tm.GetRow(0)^particleNorm));

		// Y
		tm.SetRow(1,tm.GetRow(2)^tm.GetRow(0));
	} else {
		// Choose Y in the world XY plane
		// Y
		tm.SetRow(1,Normalize(Point3(0,0,1)^tm.GetRow(0)));

		// Z
		tm.SetRow(2,tm.GetRow(0)^tm.GetRow(1));
		}

	// swap axis around	
	Point3 tmp;
	switch (GetAxis()) {
		case 1: 
			tmp = tm.GetRow(0);
			tm.SetRow(0,-tm.GetRow(1));
			tm.SetRow(1,tmp);
			break;
		case 2: 
			tmp = tm.GetRow(0);
			tm.SetRow(0,-tm.GetRow(2));
			tm.SetRow(2,tmp);
			break;
		}
		
	
	if (Bank()) {
		// Average over NUMSTEPS samples
		Point3 pt0, pt1, pt2, v0, v1;
		float cv = 0.0f;
		TimeValue st = (int)((NUMSTEPS/2+1)*tracking * GetTicksPerFrame());
		TimeValue StartTime = t - st;
		TimeValue Inc =  (st*2)/NUM_SAMPLES;

/*				
		if (!particleOb->CurveClosed(t,0)) {
			if (u+(NUMSTEPS+2)*tracking > 1.0f) u = 1.0f - (NUMSTEPS+2)*tracking;
			if (u<0.0f) u=0.0f;
			}
*/

/*		pt1 = particleOb->InterpCurve3D(t, 0, u, InterpType()) * otm;
		u += tracking;		
		pt2 = particleOb->InterpCurve3D(t, 0, u, InterpType()) * otm;
		u += tracking;		
*/

		pt1 = pobj->ParticlePosition(StartTime,id);
		StartTime += Inc;
		pt2 = pobj->ParticlePosition(StartTime,id);
		StartTime += Inc;

		for (int i=0; i<NUMSTEPS; i++) {			
			pt0 = pt1;
			pt1 = pt2;
/*
			if (!particleOb->CurveClosed(t,0) && u>1.0f) {
				//pt2 += pt2 - pt1;
				break;
			} else {
*/
//			u   = (float)fmod(u,1.0f);
//			pt2 = particleOb->InterpCurve3D(t, 0, u, InterpType()) * otm;
			pt2 = pobj->ParticlePosition(StartTime,id);
//				}
			v0 = Normalize(pt2-pt1);
			v1 = Normalize(pt1-pt0);			
			v0.z = v1.z = 0.0f; // remove Z component.

			cv += (v0^v1).z * bankAmount / Length(pt1-pt0);
			StartTime += Inc;
//			u  += tracking;			
			}	
		if (i) {
			if (GetFlip()) cv = -cv;
			switch (GetAxis()) {
				case 0: tm.PreRotateX(cv/float(i)); break;
				case 1: tm.PreRotateY(cv/float(i)); break;
				case 2: tm.PreRotateZ(cv/float(i)); break;
				}
			}
		}

	return tm;
	}


class sMyEnumProc : public DependentEnumProc 
	{
      public :
      virtual int proc(ReferenceMaker *rmaker); 
      INodeTab Nodes;              
	};

int sMyEnumProc::proc(ReferenceMaker *rmaker) 
	{ 
	if (rmaker->SuperClassID()==BASENODE_CLASS_ID)    
			{
            Nodes.Append(1, (INode **)&rmaker);                 
TSTR n(Nodes[Nodes.Count()-1]->GetName());
DebugPrint("Name %s\n",n);
			}
     return 0;              
	}


void ParticlePosition::Update(TimeValue t)
	{
//	ShapeObject *particleOb = NULL;
	ivalid = FOREVER;

	INode *particle = NULL;
	Interval iv;
//	ParticleObject *obj;
	pblock->GetValue(particle_pick,t,particle,iv);
	if (particle == NULL) return;

	ObjectState pos = particle->EvalWorldState(t);

//	Object *ob = particle.GetEvalObject();
	float u=0.0f;
	ParticleObject *obj = (ParticleObject*)pos.obj;

	
	if (obj) {
		ivalid &= obj->ObjectValidity(t);


		SimpleParticle *pobj = (SimpleParticle *) obj;
		if (!pobj->OKtoDisplay(t)) return;
		pobj->UpdateParticles(t,particle);
		if (pobj->parts.Count() ==0) return;
//		ivalid &= os.Validity(t);
		}
	else
		{
		return;
		}

	
	
//	curval = PointOnParticle(t,particleOb,ivalid);
	int id;
	pblock->GetValue(particle_ID,t,id,FOREVER);
	curval = obj->ParticlePosition(t,id);

	TimeValue age = obj->ParticleAge(t,id);
	TimeValue life = obj->ParticleLife(t,id);

	if ((age > life) || (age < 0))
		{
		curval.x = 0.0f;
		curval.y = 0.0f;
		curval.z = 0.0f;
		}


	if (Follow()) {
		curRot = Quat(CalcRefFrame(t,obj));
	} else {
		curRot.Identity();
		}
/*
	if (particleOb->CurveClosed(t,0)) {
		flags |= PARTICLEFLAG_CLOSED;
	} else {
		flags &= ~PARTICLEFLAG_CLOSED;
		}
*/
	if (ivalid.Empty()) ivalid.SetInstant(t);
	}

void ParticlePosition::GetValue(TimeValue t, void *val, Interval &valid, GetSetMethod method)
	{
	if (!ivalid.InInterval(t)) {
		Update(t);
		}
	valid &= ivalid;
			 
	if (method==CTRL_RELATIVE) {
  		Matrix3 *mat = (Matrix3*)val;
		if (Follow()) {
			curRot.MakeMatrix(*mat);
			}
		mat->SetTrans(curval);		
	} else {
		*((Point3*)val) = curval;
		}
	}

RefTargetHandle ParticlePosition::GetReference(int i)
	{
		switch (i)
		{
//			case 0:
//				return NULL;   // ref 0 is now obsolete, percent stored in ParamBlock.
//			case PARTICLEPOS_PARTICLE_REF:
//				return particle;
			case PARTICLEPOS_PBLOCK_REF:
				return pblock;
		}
		return NULL;
	}

void ParticlePosition::SetReference(int i, RefTargetHandle rtarg)
	{
		switch (i)
		{
//			case 0:
//				if (rtarg != NULL)
//					assert(0);  // ref 0 is now obsolete, no setting allowed.
//				break;
//			case PARTICLEPOS_PARTICLE_REF:
//				particle = (INode*)rtarg; break;
			case PARTICLEPOS_PBLOCK_REF:
				pblock = (IParamBlock2*)rtarg; break;
		}
	}

RefResult ParticlePosition::NotifyRefChanged(
		Interval iv, 
		RefTargetHandle hTarg, 
		PartID& partID, 
		RefMessage msg) 
	{
	switch (msg) {
		case REFMSG_CHANGE:
			ivalid.SetEmpty();
			particle_paramblk.InvalidateUI();
			break;
		}
	return REF_SUCCEED;
	}
/*
class PickParticleRestore : public RestoreObj {
	public:
		ParticlePosition *cont;
		PickParticleRestore(ParticlePosition *c) {cont=c;}
		void Restore(int isUndo) {
			if (cont->editCont == cont) {
				particle_paramblk.InvalidateUI();
				}									
			}
		void Redo() {
			if (cont->editCont == cont && cont->particle) {
				particle_paramblk.InvalidateUI();
				}
			}
		TSTR Description() { return TSTR(_T("Pick Particle")); }
	};

BOOL ParticlePosition::SetParticleNode(INode *node)
	{
	if (node->TestForLoop(FOREVER,this)==REF_SUCCEED) {
		theHold.Begin();
//		ReplaceReference(PARTICLEPOS_PARTICLE_REF,(RefTargetHandle)node);
		theHold.Put(new PickParticleRestore(this));
		theHold.Accept(GetString(IDS_RB_PICKPARTICLE));
		ivalid.SetEmpty();
		NotifyDependents(FOREVER,PART_ALL,REFMSG_CHANGE);
		return TRUE;
	} else {
		return FALSE;
		}
	}

*/
void ParticlePosition::SetTracking(float t)
	{
	float ut = ToTrackUI(t);
	pblock->SetValue(particle_smoothness, GetCOREInterface()->GetTime(), ut);
	ivalid.SetEmpty();
	NotifyDependents(FOREVER,PART_ALL,REFMSG_CHANGE);
	}

float ParticlePosition::GetTracking()
	{
	return FromTrackUI(pblock->GetFloat(particle_smoothness, GetCOREInterface()->GetTime()));
	}

void ParticlePosition::SetBankAmount(float a)
	{
	float ua = ToBankUI(a);
	pblock->SetValue(particle_bank_amount, GetCOREInterface()->GetTime(), ua);
	ivalid.SetEmpty();
	NotifyDependents(FOREVER,PART_ALL,REFMSG_CHANGE);		
	}

float ParticlePosition::GetBankAmount()
	{
	return FromBankUI(pblock->GetFloat(particle_bank_amount, GetCOREInterface()->GetTime()));
	}

void ParticlePosition::SetFollow(BOOL f)
	{
	pblock->SetValue(particle_follow, GetCOREInterface()->GetTime(), f);
	ivalid.SetEmpty();
	NotifyDependents(FOREVER,PART_ALL,REFMSG_CHANGE);
	}

void ParticlePosition::SetAllowFlip(BOOL f)
	{
	pblock->SetValue(particle_allow_upsidedown, GetCOREInterface()->GetTime(), f);
	ivalid.SetEmpty();
	NotifyDependents(FOREVER,PART_ALL,REFMSG_CHANGE);
	}
/*
void ParticlePosition::SetConstVel(BOOL cv)
	{
	pblock->SetValue(particle_constant_vel, GetCOREInterface()->GetTime(), cv);
	ivalid.SetEmpty();
	NotifyDependents(FOREVER,PART_ALL,REFMSG_CHANGE);
	}
*/

void ParticlePosition::SetBank(BOOL b)
	{
	pblock->SetValue(particle_bank, GetCOREInterface()->GetTime(), b);
	ivalid.SetEmpty();
	NotifyDependents(FOREVER,PART_ALL,REFMSG_CHANGE);	
	}

void ParticlePosition::SetFlip(BOOL onOff)
	{
	pblock->SetValue(particle_axis_flip, GetCOREInterface()->GetTime(), onOff);
	ivalid.SetEmpty();
	NotifyDependents(FOREVER,PART_ALL,REFMSG_CHANGE);	
	}

void ParticlePosition::SetAxis(int axis)
	{
	pblock->SetValue(particle_axis, GetCOREInterface()->GetTime(), axis);
	ivalid.SetEmpty();
	NotifyDependents(FOREVER,PART_ALL,REFMSG_CHANGE);
	}

int ParticlePosition::GetAxis()
	{
	return pblock->GetInt(particle_axis, 0);
//	if (flags&PARTICLEFLAG_YAXIS) return 1;
//	if (flags&PARTICLEFLAG_ZAXIS) return 2;
//	return 0;
	}

/*--------------------------------------------------------------------*/
// ParticlePosition UI

void ParticlePosition::BeginEditParams( IObjParam *ip, ULONG flags,Animatable *prev )
	{	
	this->ip = ip;
	editCont = this;
	
	particleCD.BeginEditParams(ip, this, flags, prev);
	}

void ParticlePosition::EndEditParams( IObjParam *ip, ULONG flags,Animatable *next )
	{
	editCont = NULL;
/*
	int index = aprops.FindProperty(PROPID_INTERPUI);
	if (index>=0) {
		InterpCtrlUI *ui = (InterpCtrlUI*)aprops[index];
		if (ui->hParams) {
			ip->UnRegisterDlgWnd(ui->hParams);
			ip->DeleteRollupPage(ui->hParams);			
			}
		index = aprops.FindProperty(PROPID_INTERPUI);
		if (index>=0) {
			delete aprops[index];
			aprops.Delete(index,1);
			}
		}	
	else
*/
	particleCD.EndEditParams(ip, this, flags  | END_EDIT_REMOVEUI, next);
	}

int ParticlePosition::SetProperty(ULONG id, void *data)
	{
	if (id==PROPID_INTERPUI) {		
		if (!data) {
			int index = aprops.FindProperty(id);
			if (index>=0) {
				aprops.Delete(index,1);
				}
		} else {
			InterpCtrlUI *ui = (InterpCtrlUI*)GetProperty(id);
			if (ui) {
				*ui = *((InterpCtrlUI*)data);
			} else {
				aprops.Append(1,(AnimProperty**)&data);
				}					
			}
		return 1;
	} else {
		return Animatable::SetProperty(id,data);
		}
	}

void* ParticlePosition::GetProperty(ULONG id)
	{
	if (id==PROPID_INTERPUI /*|| id==PROPID_JOINTPARAMS*/) {
		int index = aprops.FindProperty(id);
		if (index>=0) {
			return aprops[index];
		} else {
			return NULL;
			}
	} else {
		return Animatable::GetProperty(id);
		}
	}

IOResult ParticlePosition::Save(ISave *isave)
	{	
	return IO_OK;
	}


IOResult ParticlePosition::Load(ILoad *iload)
	{
	return IO_OK;
	}

